from os import environ
from os import chdir
from os import walk
from os import stat

from os.path import join
from os.path import basename
from os.path import dirname
from os.path import normpath
from os.path import isdir
from os.path import isfile

from sys import exit

from re import match

from socket import timeout

from stat import S_ISDIR

from signal import SIGWINCH
from signal import signal

from time import sleep
from time import time

from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend as crypto_default_backend

from pexpect import spawn

from digitalocean import Droplet
from digitalocean import Manager
from digitalocean import SSHKey

from paramiko import SSHClient
from paramiko import AutoAddPolicy

from config import basedir
from config import ssh_username
from config import ssh_port
from config import ssh_timeout
from config import vm_bootstrap_timeout
from config import vm_default_packages

from util import transfer_status
from util import droplet_action_wait
from util import print_status_cycle

from shell import sigwinch_passthrough
from shell import query_yes_no


if 'TOKEN' not in environ:
    print('ERROR: missing env var TOKEN (digital ocean access token)')
    exit(1)


TOKEN = environ['TOKEN']
COMMANDS = []

manager = Manager(token=TOKEN)


def command(func):
    COMMANDS.append(func.__name__)
    return func


def help():
    print("Available commands:")
    for name in sorted(COMMANDS):
        print(f'- {name}')


# created_at
# take_snapshot(self, snapshot_name, return_dict=True, power_off=False)
# restore(self, image_id, return_dict=True)
# get_snapshots(self):

@command
def add_pub_key(path, name):
    with open(path) as f:
        user_ssh_key = f.read()

    key = SSHKey(token=TOKEN,
                 name=name,
                 public_key=user_ssh_key)
    key.create()
    print(f'[+] added ssh-key: {name}')


@command
def create(name, cpu=1, ram=1, wait=True, bootstrap_vm=True):
    """Create a new vm and return its Droplet object.

    params:
    - name: string name of the vm

    optional:
    - cpu: int number of CPUs (default: 1)
    - ram: int amount of RAM in gigabyte (default: 1)
    - wait: bool whether to wait for completion
    - bootstrap_vm: bool whether to bootstrap the node

    return:
      Droplet object of the new VM
    """
    label = f'Creating vm {name} (takes several minutes)...'
    print(f'[+] {label}', end='', flush=True)

    keys = manager.get_all_sshkeys()
    droplet = Droplet(
        token=TOKEN,
        name=name,
        region='fra1',
        image='debian-9-x64',
        distro='debian',
        size_slug=f's-{cpu}vcpu-{ram}gb',
        ssh_keys=keys,
        backups=False)
    droplet.create()

    # handle create action
    actions = droplet.get_actions()
    create_action = list(filter(lambda action: action.type == 'create',
                                actions))[0]
    if wait:
        try:
            droplet_action_wait(create_action,
                                callback=print_status_cycle(labal=label),
                                timeout=vm_bootstrap_timeout)
        except timeout as e:
            print(f'\n[-] Failed, retrying: {e}')
            return create(name, cpu, ram, wait, bootstrap_vm)

    # print droplet infos
    droplet.load()
    print('done')
    print(f'{name}: {create_action.status} -> {droplet.ip_address}')

    if bootstrap_vm:
        uname = ''
        status = print_status_cycle('Booting (takes several minutes)...')
        while 'Linux' not in uname:
            status()
            try:
                uname, stderr = ssh(droplet, 'uname -s', timeout=2,
                                    print_streams=False, print_errors=False)
            except Exception:
                # TODO: ignores Permission denied (publickey)
                pass
            sleep(0.20)
        success = bootstrap(droplet)
        if not success:
            destroy(droplet, ask=False)
            droplet = create(name, cpu, ram, wait, bootstrap_vm)

    return droplet


@command
def create_challenge(challenge_path, cpu=1, ram=1, run=True):
    """Create a new bootstrapped vm with a deployed challenge and return
    its Droplet object.

    params:
    - challenge_path: string path of the challenge

    optional:
    - cpu: int number of CPUs (default: 1)
    - ram: int amount of RAM in gigabyte (default: 1)
    - run: bool wether to run the challenge container (default: True)

    return:
      Droplet object of the new VM
    """
    if challenge_path.endswith('/'):
        challenge_path = challenge_path[:-1]
    challenge_name = basename(challenge_path)
    vms = get(challenge_name)
    if vms:
        names = [vm.name for vm in vms]
        counter = max(map(int, [name[len(name) - name[::-1].index('-')]
                                if match('.+\\-[0-9]+', name) else '0'
                                for name in names]))

        challenge_name = f'{challenge_name}-{counter + 1}'
        print(f'[*] Challenge already exists as: {", ".join(names)}')
        confirm = query_yes_no(f'Create another one as {challenge_name}?')
        if not confirm:
            return

    vm = create(challenge_name, cpu, ram)
    if vm:
        challenge(vm, challenge_path, run=run)


@command
def bootstrap(name, wait=True):
    # TODO: API should only work for a single vm
    for droplet in get(name):
        with SSHClient() as client:
            client.load_system_host_keys()
            client.set_missing_host_key_policy(AutoAddPolicy)

            try:
                client.connect(droplet.ip_address,
                               port=ssh_port,
                               username=ssh_username,
                               timeout=ssh_timeout)

                # Update system
                stdin, stdout, stderr = client.exec_command(f'apt-get update && DEBIAN_FRONTEND=noninteractive apt-get upgrade -y')
                for line in stdout:
                    print(line, end='', flush=True)
                print(stderr.read().decode())

                # download install script
                stdin, stdout, stderr = client.exec_command(f'wget https://raw.githubusercontent.com/gh2o/digitalocean-debian-to-arch/debian9/install.sh -O install.sh')
                for line in stdout:
                    print(line, end='', flush=True)
                print(stderr.read().decode())

                # install Arch Linux
                options = f'--kernel_package=linux-hardened --grub_timeout=1 --extra_packages="{" ".join(vm_default_packages)}"'
                stdin, stdout, stderr = client.exec_command(f'bash install.sh {options}')
                stdin.write('wipe this droplet\n')
                for line in stdout:
                    print(line, end='', flush=True)
                print(stderr.read().decode())
            except timeout:
                print(f'ssh: connect to host {droplet.ip_address} port 22: timed out')
                continue
            except Exception as e:
                print(str(e))

        if wait:
            start = time()
            uname = ''
            status = print_status_cycle('Booting (takes several minutes)...')
            while '-hardened' not in uname:
                status()
                try:
                    uname, stderr = ssh(droplet, 'uname -r', timeout=2,
                                        print_streams=False, print_errors=False)
                    stdout, stderr = ssh(droplet, 'systemctl enable --now docker')
                except Exception:
                    pass
                sleep(0.20)

                if time() > start + vm_bootstrap_timeout:
                    print('\n[-] Failed to spawn in time, retrying...')
                    return False

        print(f'{name}: -> {droplet.ip_address}')
        return True


def challenge(vm, challenge_path, run=True):
    """Upload, build and optionally run a challenge on a VM

    params:
    - vm: string name, regex or Droplet objects
    - challenge_path: string path of the challenge
    - start: bool wether to run the challenge container
    """
    with open(f'{challenge_path}/Dockerfile') as f:
        expose = next(iter(filter(
            lambda line: line.startswith('EXPOSE '),
            f.read().splitlines()))).split()[1]

    for droplet in get(vm):
        copy_to(vm, challenge_path, '/root/ctf')
        commands = ['docker build -t ctf ctf']
        if run:
            commands.append('docker stop ctf')
            commands.append('docker rm ctf')
            commands.append(
                f'docker run -d --restart always '
                f'-p {expose}:{expose} --name ctf -it ctf')
        ssh(vm, commands)
        print(f'started challenge -> {droplet.ip_address} {expose}')


@command
def provision_ctfd(name="ctfd", backup_name="ctfd-backup"):
    # check if

    if not get(name):
        print(f"could not find vm with name {name}. Will create one.")
        create(name)

    if not get(backup_name):
        print(f"could not find backup server for vm {backup_name}. Will create one.")
        create(backup_name, bootstrap_vm=False)

    # copy the ctfd repo, and start it
    copy_to(name, "./CTFd", "CTFd")
    ssh(name, ["chmod +x /root/ctf/CTFd/docker-entrypoint.sh",
               "pacman --noconfirm --needed -Syu docker-compose",
               "docker-compose -f /root/ctf/CTFd/docker-compose.yml up -d"])

    ip_addr = get(name)[0].ip_address

    # generate keys for backup server
    key = rsa.generate_private_key(
        backend=crypto_default_backend(),
        public_exponent=65537,
        key_size=2048)

    private_key = key.private_bytes(
        crypto_serialization.Encoding.PEM,
        crypto_serialization.PrivateFormat.PKCS8,
        crypto_serialization.NoEncryption()).decode('utf8')
    public_key = key.public_key().public_bytes(
        crypto_serialization.Encoding.OpenSSH,
        crypto_serialization.PublicFormat.OpenSSH).decode('utf8')

    ssh(name, f"echo '{public_key}' >> /root/.ssh/authorized_keys")

    backup_oneliner = f'ssh -oStrictHostKeyChecking=no {ip_addr} "tar -cjf - /root/ctf/CTFd/.data | cat" > /root/backups/$(date +%Y-%m-%d-%R).tar.gz'
    commands = [f"echo '{private_key}' > /root/.ssh/id_rsa",
                f"chmod 0600 /root/.ssh/id_rsa",
                f"echo '{backup_oneliner}' > /root/backup.sh",
                "chmod +x /root/backup.sh",
                "mkdir -p /root/backups",
                # every 10 minutes
                f"echo '*/10 * * * * bash /root/backup.sh' >> /etc/crontab",
                "crontab /etc/crontab"]
    ssh(backup_name, commands)

    print(f"[*] Done. Enjoy the ctfd at {get(name)[0].ip_address} and the backup_server at {get(backup_name)[0].ip_address}")


@command
def resize(vm, cpu=None, ram=None):
    """Resize the CPU and/or memory of matching vm's

    params:
    - vm: string name, regex or Droplet objects
    - cpu: int number of CPUs (optional if ram is defined)
    - ram: int amount of RAM in gigabyte (optional if cpu is defined)
    """
    if cpu is None and ram is None:
        print('error: need at least one value for cpu or ram')
        return

    for droplet in get(vm):
        old_slug = match('\\w-(\\d+)vcpu-(\\d+)gb', droplet.size_slug)
        old_cpu = old_slug.group(1)
        old_ram = old_slug.group(1)
        new_cpu = old_cpu if cpu is None else cpu
        new_ram = old_ram if ram is None else ram
        size_slug = f's-{new_cpu}vcpu-{new_ram}gb'

        print(f'resizing {droplet.name} to {size_slug}')
        droplet.resize(size_slug)


@command
def rename(vm, name):
    """Rename a matching vm

    params:
    - vm: string name, regex or Droplet objects
    - name: string new name of the vm
    """
    droplets = get(vm)

    if not droplets:
        print(f"error: no matching vm for {name}")
        return

    names = [d.name for d in droplets]
    if len(droplets) > 1:
        print(f'error: need exactly one vm, found {", ".join(names)}')
        return

    droplet = droplets[0]
    droplet.rename(name)


@command
def destroy(vm, wait=True, ask=True):
    """Destroy matching vm's (use with care!)

    params:
    - vm: string name, regex or Droplet objects
    """
    droplets = get(vm)
    if not droplets:
        return

    names = [d.name for d in droplets]
    if ask and not query_yes_no(f'Sure you want to destroy: {", ".join(names)}'):
        return

    list(map(lambda d: d.destroy(), droplets))

    if not wait:
        return

    for droplet in droplets:
        label = f'Destroying vm {droplet.name}...'
        print(f'[+] {label}', end='', flush=True)
        # handle destroy action
        actions = droplet.get_actions()
        destroy_action = list(filter(lambda action: action.type == 'destroy',
                                     actions))[0]
        droplet_action_wait(destroy_action,
                            callback=print_status_cycle(labal=label))


@command
def get(vm=None):
    """Get Droplet objects of matching vm's

    params:
    - vm: string name, regex or Droplet objects

    returns a list of Droplet objects
    """
    if isinstance(vm, Droplet):
        return [vm]

    if isinstance(vm, list) and all([isinstance(v, Droplet) for v in vm]):
        return vm

    droplets = manager.get_all_droplets()
    if not vm:
        return droplets
    return list(filter(lambda d: match(f'^{vm}$', d.name), list(droplets)))


@command
def ls(vm=None):
    """List details of matching vm's

    params:
    - vm: string name, regex or Droplet objects
    """
    fallback_ip = 'unassigned'

    droplets = sorted(get(vm), key=lambda d: d.name)
    name_length = max([len(d.name) for d in droplets]) + 1
    ip_length = max([len(d.ip_address or fallback_ip) for d in droplets]) + 2
    status_length = max([len(d.status) for d in droplets]) + 1
    header = 'name'.ljust(name_length) + \
             'ip'.rjust(ip_length // 2 + 2).ljust(ip_length) + ' ' + \
             'cpu '.rjust(4) + \
             'mem '.rjust(6) + \
             'hdd '.rjust(4) +\
             'status'.rjust(status_length)
    print('=' * len(header))
    print(header)
    print('=' * len(header))
    for droplet in droplets:
        ip = droplet.ip_address or fallback_ip
        print(f'{droplet.name:<{name_length}}'
              f'{ip:>{ip_length}} '
              f'{droplet.vcpus:>3} '
              f'{droplet.memory:>5} '
              f'{droplet.disk:>3} '
              f'{droplet.status:>{status_length}}')
    print('-' * len(header))


@command
def power_on(vm):
    """Power on all matching vm's

    params:
    - vm: string name, regex or Droplet objects
    """
    list(map(lambda d: d.power_on(), get(vm)))


@command
def power_off(vm):
    """Power off all matching vm's

    params:
    - vm: string name, regex or Droplet objects
    """
    list(map(lambda d: d.power_off(), get(vm)))


@command
def power_cycle(vm):
    """Power cycle all matching vm's

    params:
    - vm: string name, regex or Droplet objects
    """
    list(map(lambda d: d.power_cycle(), get(vm)))


@command
def shutdown(vm):
    """Shutdown all matching vm's

    params:
    - vm: string name, regex or Droplet objects
    """
    list(map(lambda d: d.shutdown(), get(vm)))


@command
def reboot(vm):
    """Reboot all matching vm's

    params:
    - vm: string name, regex or Droplet objects
    """
    list(map(lambda d: d.reboot(), get(vm)))


@command
def status(vm):
    """Print status entries on all matching vm's

    params:
    - vm: string name, regex or Droplet objects
    """
    for droplet in get(vm):
        print(f'{droplet.name}: {droplet.status}')
        actions = droplet.get_actions()
        max_type = max(max([len(action.type) for action in actions]) + 2, 12)
        for action in actions:
            print(f' - {action.completed_at or action.started_at} {action.type:>{max_type}}: {action.status}')
        print()


@command
def ssh(vm, command=None, print_streams=True, print_errors=True, timeout=ssh_timeout):
    """Execute passed shell commands or invoke an interactive shell on all
    matching vm's

    params:
    - vm: string name, regex or Droplet objects

    optional:
    - command: empty to pop interactive shell
               string to execute a single commands
               list of strings to execute multiple commands
    """
    stdout_all = ''
    stderr_all = ''
    for droplet in get(vm):

        if not command:
            child = spawn(f'ssh -p {ssh_port} {ssh_username}@{droplet.ip_address}',
                          encoding='utf-8', timeout=timeout)
            signal(SIGWINCH, sigwinch_passthrough(child))
            child.interact()
            continue

        with SSHClient() as client:
            client.load_system_host_keys()
            client.set_missing_host_key_policy(AutoAddPolicy)

            try:
                client.connect(droplet.ip_address,
                               port=ssh_port,
                               username=ssh_username,
                               timeout=timeout)

                def exec(command):
                    stdin, stdout, stderr = client.exec_command(command)
                    stdout_all = ''
                    stderr_all = ''
                    for line in stdout:
                        if print_streams:
                            print(line, end='', flush=True)
                        stdout_all += line
                    for line in stderr:
                        if print_streams:
                            print(line, end='', flush=True)
                        stderr_all += line
                    return stdout_all, stderr_all

                if isinstance(command, str):
                    stdout, stderr = exec(command)
                    stdout_all += stdout
                    stderr_all += stderr
                    continue

                if isinstance(command, list):
                    stdouts, stderrs = list(zip(*map(exec, command)))
                    stdout = '\n'.join(stdouts)
                    stderr = '\n'.join(stderrs)
                    stdout_all += stdout
                    stderr_all += stderr
                    continue

                # channel = client.invoke_shell()
                # posix_socket_shell(channel)
            except timeout:
                if print_errors:
                    print(f'ssh: connect to host {droplet.ip_address} port 22: timed out')

    if not print_streams:
        return stdout_all, stderr_all


@command
def copy_to(vm, copy_from, copy_to):
    """Copies a file or a directory to the target VM's

    The CWD on the local host is the root dir of the challenge repository
    The CWD on the target host is /root/ctf

    params:
    - vm: string name, regex or Droplet objects
    - copy_from: string a relative or absolute path (file or directory)
    - copy_to: string a relative or absolute path (file or directory)
    """

    ctf_working_dir = '/root/ctf'
    copy_from = normpath(copy_from)
    copy_to = normpath(copy_to)
    chdir(basedir)

    for droplet in get(vm):
        with SSHClient() as client:
            client.load_system_host_keys()
            client.set_missing_host_key_policy(AutoAddPolicy)

            try:
                client.connect(droplet.ip_address,
                               port=ssh_port,
                               username=ssh_username,
                               timeout=ssh_timeout)

                with client.open_sftp() as sftp:
                    # ensure default ctf cwd exists and is set
                    if not basename(ctf_working_dir) in sftp.listdir(dirname(ctf_working_dir)):
                        sftp.mkdir(ctf_working_dir, 0o755)
                    sftp.chdir(ctf_working_dir)

                    if isdir(copy_from):
                        # ensure target directory exists
                        if not basename(copy_to) in sftp.listdir(dirname(copy_to)):
                            sftp.mkdir(copy_to, 0o755)

                        # recursive directory copy
                        for root, dirs, files in walk(copy_from, topdown=True):
                            for name in files:
                                f_from = join(root, name)
                                f_to = f_from.replace(copy_from, copy_to, 1)
                                sftp.put(f_from, f_to, callback=transfer_status(f_to))
                                sftp.chmod(f_to, stat(f_from).st_mode)
                                print()
                            for name in dirs:
                                d = join(root, name).replace(copy_from, copy_to, 1)
                                if not basename(d) in sftp.listdir(dirname(d)):
                                    sftp.mkdir(d, stat(copy_from).st_mode)

                    elif isfile(copy_from):
                        # check if target is an existing directory to copy into
                        if basename(copy_to) in sftp.listdir(dirname(copy_to)):
                            to_stat = sftp.stat(copy_to)
                            if S_ISDIR(to_stat.st_mode):
                                copy_to = join(copy_to, basename(copy_from))
                        sftp.put(copy_from, copy_to, callback=transfer_status(copy_to))
                        sftp.chmod(copy_to, stat(copy_from).st_mode)
                        print()

            except timeout:
                print(f'ssh: connect to host {droplet.ip_address} port 22: timed out')


@command
def update(vm, reboot=None):
    for droplet in get(vm):
        reboot_message = f'[*] Kernel of {droplet.name} upgraded, rebooting...'
        commands = [
            f'pacman --noconfirm --needed -Syu {" ".join(vm_default_packages)}'
        ]
        # auto reboot detection on kernel
        if reboot is None:
            commands.append('[[ "$(uname -r|sed -E "s|(.+)-hardened|\\1|")" = '
                            '   "$(pacman -Qs "^linux-hardened$"|head -1|cut -d" " -f2)" ]] '
                            f' || (echo "{reboot_message}" && systemctl reboot)')
        if reboot:
            print(reboot_message)
            commands.append('systemctl reboot')
        ssh(droplet, commands)
